# 10.02.2012, End of Line: UNIX

# README: The following functions were written in order to perform the tests 
# suggested and/or discussed in the paper. Note that using these functions 
# for large sample sizes and/or for a large number of permutations 
# may take a substantial amount of computational time. Faster 
# versions of the functions are in development. 

# We provide also the test procedures of 
#   Anderson, M.J. (2006) Distance-based tests for homogeneity of
#   multivariate dispersions. _Biometrics_ *62(1)*, 245-253.
# as the current implementation in 'vegan' package is not 
# in agreement with the software that is available at 
# http://www.stat.auckland.ac.nz/~mja/Programs.htm 



# # F_{d} or F_{d}^{log} test
Fd.test <- function(mat, ident, n.perm = 999, Fdlog = FALSE, input="distances", centring="median"){
# 	mat ... a matrix of pairwise distances (either a numeric matrix or an object of clas 'dist') or a data matrix with numeric observations in rows 
# 	ident ... group identifier
# 	n.perm ... number of permutation
# 	Fdlog = TRUE ... Fdlog version of the test will be used
# 	input ... "distance" for a distance matrix, "data" for  a data matrix. In the latter case the Euclidean distance is used. For other dissimilarity  measures insert a matrix of distances. 
# 	centring ... "median" for a spatial median, "centroid" for a centroid (a component-wise mean) 

	# Checks for input
	if(sum(is.element(input, c("data", "distances"))) == 0){
		print("Unknown type of input.");
		print("Specify the input either as 'data' or 'distances'. ");
		stop();
	}
	
	if(sum(is.element(centring, c("centroid", "median"))) == 0){
		print("Unknown type of centring.");
		print("Specify the input either as 'centroid' or 'median'. ");
		stop();
	}
	
	if(input == "data"){
		X <- mat;
		mat <- dist(mat);
	}
	
  mat <- as.matrix(mat);
	
	# Checks for input
	if(nrow(mat) != length(ident)){
		print("The number of the rows of the 'input' does not match with the length of the 'ident'.");
		stop();
	}
	
	if(nrow(mat) != ncol(mat)){
		print("The 'input' parameter cannot be a matrix of distances, as the number of rows and columns differ.");
		stop();
	}
	
	
	ident.orig <- ident;
	ident <- as.numeric(as.factor(ident));
		
  hodnoty.ident <- unique(ident);
	
	if(length(hodnoty.ident) <= 1){
		print("There is only one value in 'ident' variable.");
		print("The groups cannot be identified.");
		stop();
	}
  
  nn <- as.numeric(table(ident));
  n <- sum(nn);
  K <- length(nn);
	df0 <- n - K;
#   browser();
	
  qqq <- outer(ident, ident, "!=");
  matB <- mat;
	matB[qqq] <- 0;	
#  The  mean distance of the i-th observations from the others
  Di <- rowSums(matB)/(nn[ident]-1);  
# 	Mean group distance 
  mu <- tapply(Di, ident, mean);  
	sigma <- 4*(nn-1)^2/(nn-2)^2 * tapply(Di, ident, var);
	
	mu.orig <- mu;
	sigma.orig <- sigma;
	Di.orig <- Di;
	
# #  Taking log of the mean within distances
  if(Fdlog == TRUE){
    sigma <- sigma/mu^2;
    mu <- log(mu);
    sh2 <- sum(sigma*(nn-1))/(sum(nn)-K);
    mu.all <- sum(mu*(nn))/sum(nn);
# 	Test statistics
    stat <- sum((nn)*(mu - mu.all)^2)/(K - 1)/sh2;
  }else{
	  sh2 <- sum(sigma*(nn-1))/(sum(nn)-K);
		mu.all <- sum(mu*(nn))/sum(nn);
# 	Test statistics
		stat <- sum((nn)*(mu - mu.all)^2)/(K - 1)/sh2;
	}
	
	  if(n.perm <= 0){
			DO.PERM <- FALSE;
		}else{
			DO.PERM <- TRUE;
		}
	 
# # 	The input is the raw dataset
	if(input == "data"){
		podil.neg <- 0;
		n.neg <- 0;
		
		if(centring == "centroid"){
			DATAc <- lm(X~as.factor(ident))$res;
		}
		if(centring == "median"){
			spatmed <- ordimedian.pcaPP(X, ident);
			DATAc <- X - spatmed[ident,];		
		}
		

# 	Distance matrix from the centred data
			matc <- as.matrix(dist(DATAc));
	}
	
	p.val.perm <- NA;
	
  if(input == "distances"){
		
		fit <- try(betadisper2(as.dist(mat), ident, type = centring));
		
		if(class(fit) != "betadisper"){
			warning("PCoA has failed. Only asymptotic version of the test is computed.");
			DO.PERM <- FALSE;
# 			browser();
		}else{
			n.neg <- sum(!fit$pos);
			
	# # 		Are imaginary axes are present in the representation?
			if(n.neg > 0){
	# 			browser();
	# 			dblcen <- function(A) A - outer(rowMeans(A), colMeans(A), "+")  + mean(A);
				
					# #   Correction to prevent imaginary axes in the representation 
				A <- -0.5*mat^2;
# 				storage.mode(A) <- "double"
# 				.C("dblcen", A, as.integer(n), DUP = FALSE, PACKAGE="stats");
# 				D1 <- A;
				D1 <- dblcen(A);
				
				A <- -0.5*mat;
# 				storage.mode(A) <- "double"
# 				.C("dblcen", A, as.integer(n), DUP = FALSE, PACKAGE="stats");
# 				D2 <- A;
				D2 <- dblcen(A);
				
				A <- rbind(cbind(diag(0, n), 2*D1), cbind(diag(-1, n), -4*D2));
				const2 <- max(as.numeric(sort(eigen(A, symmetric = FALSE, only.values=TRUE)$values)[2*n]),0);
				
				fitp <- try(betadisper2(as.dist(mat+const2), ident, type = centring));      
# 				browser()
			# 		afit <- anova(fit);       
				
			# Was the correction method succesful?
				q.repr <- (length(grep("Error", fitp)) != 0) | (sum(as.logical(fitp$eig < -0.0001)) > 0);
				i0 <- 0;
				
				while((q.repr == TRUE) & i0 < 3){
					i0 <- i0 + 1;
					print(paste("Corrected distance matrix has negative eigenavalues:", i0));
					
					fitp <- try(betadisper2(as.dist(mat+const2), ident, type = "median"));
					q.repr <- (length(grep("Error", fitp)) != 0) | (sum(as.logical(fitp$eig < -0.0001)) > 0);        
				}
				
			
			#  matc ... the centred matrix
				if(q.repr){
					pos <- as.logical(fit$eig >= 0);
					matc <- sqrt(abs(as.matrix(dist(fit$vectors.cent[, pos]))^2 - as.matrix(dist(fit$vectors.cent[, !pos]))^2));
				}else{
					matc <- as.matrix(dist(fitp$vectors.cent) - const2);
# 					browser();
				}
			}else{
				matc <- as.matrix(dist(fit$vectors.cent));
			}
		}
	}
  
	# # #   Permutations 
	stat.perm <- rep(NA, n.perm);  
	
	if(DO.PERM){
		for(j in 1:n.perm){
			indexy <- sample(n);
			indexy.perm <- ident[indexy]

			matP <- matc[indexy, indexy];# permuted centred matrix

			matP[qqq] <- 0;	
		#  The  mean distance of the i-th observations from the others
			Di <- rowSums(matP)/(nn[ident]-1);  
		# 	Mean group distance 
			mu <- tapply(Di, ident, mean);  
			sigma <- 4*(nn-1)^2/(nn-2)^2 * tapply(Di, ident, var);
			
		# #  Taking log of the mean within distances
			if(Fdlog == TRUE){
				sigma <- sigma/mu^2;
				mu <- log(mu);
				sh2 <- sum(sigma*(nn-1))/(sum(nn)-K);
				mu.all <- sum(mu*(nn))/sum(nn);
				stat.perm[j] <- sum((nn)*(mu - mu.all)^2)/(K - 1)/sh2;
			}else{
				sh2 <- sum(sigma*(nn-1))/(sum(nn)-K);

				mu.all <- sum(mu*(nn))/sum(nn);
				stat.perm[j] <- sum((nn)*(mu - mu.all)^2)/(K - 1)/sh2;
			}
		}
	# #   Permutation test		
			p.val.perm <- (sum(round(stat.perm,9) >= round(stat,9))+1)/(n.perm + 1);
	}
# #   Calculating p-values
#  F-distrubion
  p.val <- (1 - pf(stat, df1 = K-1, df2 = df0));
	
# 	hodnoty.ident <- levels(as.factor(ident.orig));
	hodnoty.ident <- levels(as.factor(ident.orig))[table(ident.orig) > 0];
	
	print(paste("Sample sizes: "));
	names(nn) <- hodnoty.ident;
	print(nn);	
	print(paste("Logarithm of mean within distances was used in test statistic:", Fdlog));
  print(paste("Test statistics:", round(stat,4)));
	print(paste("P-value (F-distribution):", round(p.val,5)));
	print(paste("P-value (permuation):", round(p.val.perm,5)));
	print(paste("Number of permutations:", n.perm));
	print(paste("Estimator used for centring: ", centring));
	print("")
		
	print("Mean within distances:");
	names(mu.orig) <- hodnoty.ident;
  print(mu.orig);
  print("");
	

	l <- list();
	l$groups <- hodnoty.ident;
	l$Fstat <- stat;
	l$mu <- mu.orig;
	names(sigma.orig) <- hodnoty.ident;
	l$sigma <- sigma.orig;
	l$Di <- Di.orig;
	l$ident <- ident.orig;
	
	invisible(l)
}


# # Tests F_{And}  of Anderson (2006)
FAnd.test <- function(mat, ident, n.perm = 999, input="distances", centring="median"){
# 	mat ... a matrix of pairwise distances (either a numeric matrix or an object of clas 'dist') or a data matrix with numeric observations in rows 
# 	ident ... group identifier
# 	n.perm ... number of permutation
# 	input ... "distance" for a distance matrix, "data" for  a data matrix. In the latter case the Euclidean distance is used. For other dissimilarity  measures insert a matrix of distances. 
# 	centring ... "median" for a spatial median, "centroid" for a centroid (a component-wise mean) 

# Checks for input
	if(sum(is.element(input, c("data", "distances"))) == 0){
		print("Unknown type of input.");
		print("Specify the input either as 'data' or 'distances'. ");
		stop();
	}
	
	if(sum(is.element(centring, c("centroid", "median"))) == 0){
		print("Unknown type of centring.");
		print("Specify the input either as 'centroid' or 'median'. ");
		stop();
	}
	
	if(input == "data"){
		X <- as.matrix(mat);
		mat <- dist(mat);
	}
	
  mat <- as.matrix(mat);
	
	if(nrow(mat) != length(ident)){
		print("The number of the rows of the 'input' does not match with the length of the 'ident'");
		stop();
	}
	
	
	if(nrow(mat) != ncol(mat)){
		print("The 'input' parameter cannot be a matrix of distances, as the number of rows and columns differ");
		stop();
	}
	
	
	ident.orig <- ident;
	ident <- as.numeric(as.factor(ident));
		
  hodnoty.ident <- unique(ident);
  
  nn <- as.numeric(table(ident));
  n <- sum(nn);
  K <- length(nn);
	df0 <- n - K;
  
# # 	The input is a dataset
	if(input == "data"){
		podil.neg <- 0;
		n.neg <- 0;
		
		fit <- list();
		fit$vectors <- X;
		fit$pos <- rep(TRUE, ncol(X));
		
		if(centring == "centroid"){			
			fit$vectors.cent <- lm(X~as.factor(ident))$res;
			fit$distances <- sqrt(rowSums(as.matrix(fit$vectors.cent)^2)); 
		}
		if(centring == "median"){
			spatmed <- ordimedian.pcaPP(X, ident);
			fit$vectors.cent <- X - spatmed[ident,];
			fit$distances <- sqrt(rowSums(as.matrix(fit$vectors.cent)^2));
		}		
	}
	
# # 	The input is a distance matrix
  if(input == "distances"){
		
		fit <- betadisper2(as.dist(mat), ident, type = centring);
		
		n.neg <- sum(!fit$pos);
		if(n.neg > 0){
# 			podil.neg <- (-1)*sum(fit$eig[!fit$pos])/sum(fit$eig[fit$pos]);
# 	  browser();
			fit$distances <- sqrt(abs(rowSums(as.matrix(fit$vectors.cent[,fit$pos])^2) - rowSums(as.matrix(fit$vectors.cent[,!(fit$pos)])^2)));
		}else{
# 			podil.neg <- 0;
			fit$distances <- sqrt(rowSums(fit$vectors.cent^2));
		}
	}
  
# QR-decomposition to speed up computation
  mod <- lm(fit$distances ~ as.factor(ident));
  mod.Q <- mod$qr
  p <- mod.Q$rank
  stat <- summary(mod)$fstatistic[1];
  
# # #   Permutations 
  stat.perm <- rep(NA, n.perm);  
	
  for(j in 1:n.perm){
		perm <- sample(n);
		X <- fit$vectors.cent[perm,];

		if(centring == "centroid"){
			X <- qr.resid(mod.Q, X)
		}else{
			spatmed <- ordimedian.pcaPP(X, ident);
			X <- X - spatmed[ident,];
		}
		
		perm.resid <- sqrt(abs(rowSums(X[,fit$pos]^2) - rowSums(X[,!fit$pos]^2)));
    f <- qr.fitted(mod.Q, perm.resid)
    mss <- sum((f - mean(f))^2)/(K-1);
		r <- perm.resid - f;
    rss <- sum(r^2)
        
    resvar <- rss / df0
    stat.perm[j] <- mss/resvar;        
  }
  
# #   Calculating p-values
#  F-distrubion
  p.val <- (1 - pf(stat, df1 = K-1, df2 = df0));
	
# #   Permutation test
  p.val.perm <- (sum(round(stat.perm,9) >= round(stat,9))+1)/(n.perm + 1);
	
	hodnoty.ident <- levels(as.factor(ident.orig))[table(ident.orig) > 0];
			
	print(paste("Sample sizes: "));
# 	print(table(ident.orig));	
	names(nn) <- hodnoty.ident;
	print(nn);	
  print(paste("Test statistics:", round(stat,4)));
	print(paste("P-value (F-distribution):", round(p.val,5)));
	print(paste("P-value (permuation):", round(p.val.perm,5)));
	print(paste("Number of permutations:", n.perm));
	print(paste("Estimator used for centring: ", centring));
	print("")
		
	print("Mean distances from the centre: ");
  print(tapply(fit$distances, ident.orig, mean));

	l <- list();
	l$Fstat <- stat;
	l$perm.stat <- stat.perm;
	
	invisible(l)
}


# # A modified 'betadisper' function from the package "vegan" 
# http://CRAN.R-project.org/package=vegan
betadisper2 <- function(d, group, type = c("median","centroid"))
{
    ## inline function for spatial medians
    spatialMed <- function(vectors, group, pos) {
#         axes <- seq_len(NCOL(vectors))
#         spMedPos <- ordimedian(vectors, group)
        spMedPos <- ordimedian.pcaPP(vectors, group)
#         spMedNeg <- ordimedian(vectors, group, choices = axes[!pos])
        cbind(spMedPos)
    }
    ## inline function for distance computation
    Resids <- function(x, c) {
        if(is.matrix(c))
            d <- x - c
        else
            d <- sweep(x, 2, c)
        rowSums(d^2)
    }
    ## Tolerance for zero Eigenvalues
    TOL <- 1e-7
    ## uses code from stats:::cmdscale by R Core Development Team
    if(!inherits(d, "dist"))
        stop("distances 'd' must be a 'dist' object")
    ## Someone really tried to analyse correlation like object in range -1..+1
    if (any(d < -TOL, na.rm = TRUE))
        stop("dissimilarities 'd' must be non-negative")
    if(missing(type))
        type <- "median"
    type <- match.arg(type)
    ## checks for groups - need to be a factor for later
    if(!is.factor(group))
        group <- as.factor(group)
    n <- attr(d, "Size")
    x <- matrix(0, ncol = n, nrow = n)
    x[row(x) > col(x)] <- d^2
    ## site labels
    labs <- attr(d, "Labels")
    ## remove NAs in group
    if(any(gr.na <- is.na(group))) {
        group <- group[!gr.na]
        x <- x[!gr.na, !gr.na]
        ## update n otherwise C call crashes
        n <- n - sum(gr.na)
        ## update labels
        labs <- labs[!gr.na]
        warning("Missing observations due to 'group' removed.")
    }
    ## remove NA's in d
    if(any(x.na <- apply(x, 1, function(x) any(is.na(x))))) {
        x <- x[!x.na, !x.na]
        group <- group[!x.na]
        ## update n otherwise C call crashes
        n <- n - sum(x.na)
        ## update labels
        labs <- labs[!x.na]
        warning("Missing observations due to 'd' removed.")
    }
    x <- x + t(x)
    storage.mode(x) <- "double"
#     .C("dblcen", x, as.integer(n), DUP = FALSE, PACKAGE="stats");
#     dblcen <- function(A) A - outer(rowMeans(A), colMeans(A), "+")  + mean(A);
    x <- dblcen(x);
#     browser();   
    
    x <- (x+t(x))/2;
    e <- eigen(-x/2, symmetric = TRUE);
		
# #  Checking the PCoA representation. If not switching from LAPACK to EISPACK.
    if(sum(is.na(e$values)) + sum(is.na(e$vectors)) > 0) {
			warning("LAPACK routines to find eigenvalues failed. Using EISPACK.");
      e <- eigen(-x/2, symmetric = TRUE, EISPACK = TRUE);
    }else{
      if(max(e$values) > 1e+6){
        warning("LAPACK routines to find eigenvalues failed. Using EISPACK.");
        e <- eigen(-x/2, symmetric = TRUE, EISPACK = TRUE);
#         browser();
      }
    }
    
#     e <- eigen(-x/2, symmetric = TRUE)
    vectors <- e$vectors
    eig <- e$values
    ## Remove zero eigenvalues
    eig <- eig[(want <- abs(eig/eig[1]) > TOL)]
    ## scale Eigenvectors
    if(length(eig) == 1){
			vectors <- vectors[, want, drop = FALSE] * sqrt(abs(eig));
			if(type == "centroid"){
				centroids <- tapply(vectors, group,mean);
				zij <- abs(resid(lm(vectors~group)));
			}else{
				centroids <- tapply(vectors, group, median);
# 				browser();
				zij <- abs(as.numeric(vectors) - centroids[group]);
			}
		}else{
			vectors <- vectors[, want, drop = FALSE] %*% diag(sqrt(abs(eig)));
    ## store which are the positive eigenvalues
    pos <- eig > 0
#     pos <- abs(eig) > 0;
    ## group centroids in PCoA space
    centroids <-
        switch(type,
               centroid = apply(vectors, 2, function(x) tapply(x, group, mean)),
               median = spatialMed(vectors, group, pos)
               )
    ## for each of the groups, calculate distance to centroid for
    ## observation in the group
    ## Uses in-line Resids function as we want LAD residuals for
    ## median method, and LSQ residuals for centroid method
    dist.pos <- Resids(vectors[, pos, drop=FALSE],
                       centroids[group, pos, drop=FALSE])
    dist.neg <- 0
    if(any(!pos))
        dist.neg <- Resids(vectors[, !pos, drop=FALSE],
                           centroids[group, !pos, drop=FALSE])

    ## zij are the distances of each point to its group centroid
    zij <- sqrt(abs(dist.pos - dist.neg))
		
    ## add in correct labels
    colnames(vectors) <- names(eig) <- paste("PCoA", seq_along(eig), sep = "")
		vectors.cent <- vectors - centroids[group,];
# 		browser();
    if(is.matrix(centroids))
        colnames(centroids) <- names(eig)
    else
        names(centroids) <- names(eig)
    rownames(vectors) <- names(zij) <- labs
		}
    retval <- list(eig = eig, vectors = vectors, vectors.cent = vectors.cent, pos = pos, distances = zij, type = type, 
                   group = group, centroids = centroids, call = match.call())
    class(retval) <- "betadisper"
    attr(retval, "method") <- attr(d, "method")
    attr(retval, "type") <- type
    retval
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# The following function calculates spatial medians in groups given by 'groups' argument. 
# This function requires the function 'l1median_NLM' from library(pcaPP). 
# If this library is not available, the function 'l1median_NLM' has to be replaced. 

library(pcaPP);

ordimedian.pcaPP <- function(y, groups){
  
  inds <- names(table(groups));
  n.inds <- length(inds); 
  medians <- matrix(0, nrow = n.inds, ncol = ncol(y))
  rownames(medians) <- inds
#   colnames(medians) <- colnames(pts)
  for (i in 1:n.inds) {
    X <- y[groups == inds[i], , drop = FALSE]
    medians[i, ] <- l1median_NLM(X)$par;
  }
                      
#   invisible(medians)
  medians;
}

dblcen <- function(x, na.rm = TRUE) {
  cnt <- colMeans(x, na.rm = na.rm)
  x <- sweep(x, 2L, cnt, check.margin = FALSE)
  cnt <- rowMeans(x, na.rm = na.rm)
  sweep(x, 1L, cnt, check.margin = FALSE)
}